package nz.ac.massey.javaecs;
/**
 * Component Manager
 * This class manages component registrations, and adding,
 * removing, and retrieving data from the component arrays.
 * 
 * Contributors:
 * Brychan Dempsey - brychand@hotmail.com
 * 
 */
import java.util.Map;
import java.lang.reflect.Type;
import java.util.BitSet;
import java.util.HashMap;

/**
 * Manages the addition, sorting and retrieving of components and component data
 */
public class ComponentManager{
    private Map<Type, Map<Entity, Object>> componentArrays = new HashMap<>();
    // Need to be able to map bit indices and component types
    private Map<Integer, Type> indexComponentType = new HashMap<>();
    private Map<Type, Integer> componentTypeIndex = new HashMap<>();
    // 

    /**
     * Adds the specified component to the provided entity.
     * <p>
     * Providing a null Object defaults the value to boolean false
     * @param componentType the class type of the component to add
     * @param componentData the component data to associate
     * @param entity the entity to associate data to
     */
    protected boolean addComponentToEntity(Type componentType, Object componentData, Entity entity){
        componentArrays.get(componentType).put(entity, componentData);
        return true;
    }

    /**
     * Signals to the ComponentManager the entity was destroyed. All component data references should be removed.
     * @param entity the entity that was destroyed.
     */
    public void entityDestroyed(Entity entity, BitSet entityRegistrations){
        // HashMap lookups take time, use the known bitstates to avoid
        int index = entityRegistrations.nextSetBit(0);
        while(index != -1){
            componentArrays.get(indexComponentType.get(index)).remove(entity);
            index = entityRegistrations.nextSetBit(index+1);
        }
    }

    /**
     * Gets the component data associated with the entity
     * @param componentType the class type of data to look for
     * @param entity the entity to find data for
     * @return the Object data found, or null if it was not found
     */
    public Object getComponent(Type componentType, Entity entity){
        return componentArrays.get(componentType).get(entity);
    }

    /**
     * Gets the registration index of the component type
     * @param type the class type of the component
     * @return the index of the component type, or -1 if it isn't found
     */
    public Integer getComponentIndex(Type type){
        try{
            return componentTypeIndex.get(type);
        }
        catch (NullPointerException e){
            return -1;
        }        
    }

    /**
     * Gets the type of the component at the provided index
     * @param index the index of the component
     * @return the class type of the index. `null` if not found
     */
    public Type getComponentType(Integer index){
        try{
            return indexComponentType.get(index);
        }
        catch (NullPointerException e){
            return null;
        }  
    }

    /**
     * Moves a single component data from one entity to another
     * @param sourceEntity the entity to move data from
     * @param destinationEntity the entity to move data to
     * @param component the component class type to consider 
     * @return true if the component was moved successfully, else false
     */
    public boolean moveComponentData(Entity sourceEntity, Entity destinationEntity, Type component){
        Object data = componentArrays.get(component).get(sourceEntity);
        componentArrays.get(component).put(destinationEntity, data);
        componentArrays.get(component).remove(sourceEntity);
        return true;
    }

    /**
     * Moves all component data from one entity to another
     * @param sourceEntity the entity to move data from
     * @param destinationEntity the entity to move data to
     * @param sourceRegistrations the component registrations of the source entity
     */
    public void moveAllComponentData(Entity sourceEntity, Entity destinationEntity, BitSet sourceRegistrations){
        int result = sourceRegistrations.nextSetBit(0);
        while (result != -1){
            moveComponentData(sourceEntity, destinationEntity, indexComponentType.get(result));
            result = sourceRegistrations.nextSetBit(result +1);
        }
    }

    /**
     * Registers the component type
     * @param type the class type to register
     * @return true if the component was registered successfully, else false
     */
    public boolean registerComponent(Type type){
        return registerComponent(type, 16);
    }

    /**
     * Registers the component type
     * @param type the class type to register
     * @param arraySize the number of elements to prereserve space for.
     * @return true if the component was registered successfully, else false
     */
    public boolean registerComponent(Type type, int arraySize){
        if (componentArrays.containsKey(type)){
            Engine.writeErr("Component " + type.getTypeName() + " is already registered");
            return false;
        }
        componentArrays.put(type, new HashMap<>());
        indexComponentType.put(componentTypeIndex.size(), type);
        componentTypeIndex.put(type, componentTypeIndex.size());
        return true;
    }

    /**
     * Removes the specified component from the entity.
     * <p>
     * Be sure to call SystemManager.entityRegistrationsChanged after calling this function
     * @param componentType the class type of the component to remove
     * @param entity the entity to remove the component from
     */
    public boolean removeComponentFromEntity(Type componentType, Entity entity){
        componentArrays.get(componentType).remove(entity);
        return true;
    }

    /**
     * Checks if the entity contains data for the provided component
     * @param entity the entity to check
     * @param componentType the component class type
     * @return true if the entity has component data
     */
    public boolean entityHasComponentData(Entity entity, Type componentType){
        if (componentArrays.get(componentType).containsKey(entity)) return true;
        else return false;
    }
}