package nz.ac.massey.javaecs;
/**
 * ECS System class
 * This class stores entity-system registrations. These are
 * updated by SystemManager, but need a type to store the
 * data under.
 * 
 * All systems should inherit from this
 * 
 * Contributors:
 * Brychan Dempsey - brychand@hotmail.com
 * 
 */

import java.util.Set;
import java.util.BitSet;
import java.util.HashSet;

/**
 * Abstract class that all systems should inherit from<p>
 * Defines four required components:
 * <ol>
 * <li>The list of <b>entities</b> that have all the components required for this system</li>
 * <li>The <b>BitSet</b> of component registrations required for this system</li>
 * <li>The <b>init()</b> function, where logic that needs to be run once is performed</li>
 * <li>The <b>update(dt)</b> function, where logic that needs to be run regularly is performed. <i>dt is the delta time in millseconds</i></li>
 * </ol>
 * Additionally, the object constructor should define the actual values of the BitSet.<p>
 * Additional functions can be implemented as required.
 */
public abstract class ECSSystem{
    protected Set<Entity> entities = new HashSet<>();
    protected BitSet registrationSet;

    public ECSSystem(){
        registrationSet = new BitSet();
    }

    public BitSet getRegistrationSet(){
        return registrationSet;
    }

    /**
     * Functionality that should be run only once.
     * Implement additional parameterised <b>init()</b> functions as required.
     * <p>
     * This is distinct from the constructor as this may be run as required any time between
     * the construction and <b>update(dt)</b>. It is not intended to set the actual bits of <b>registrationSet</b>
     * in this function; that should be performed in the constructor.
     * @throws Exception can return exceptions to the main program if an issue occurred. 
     */
    public abstract void init() throws Exception;
    
    /**
     * Functionality that is expected to be called regularly
     * Intended as a template only; may be superficially implemented.
     * Implement additional parameterised <b>update()</b> functions as required.
     * @param dt delta-time; the change in time in milliseconds since this function was last run
     */
    public abstract void update(double dt);
}