package nz.ac.massey.javaecs;
/**
 * ECS engine class.
 * Call this class and its functions to interact correctly with the ECS system.
 * This is the entry-point to the library
 * 
 * Contributors:
 * Brychan Dempsey - brychand@hotmail.com
 * 
 * References:
 * Based on the implementation by Austin Morlan:
 * https://austinmorlan.com/posts/entity_component_system/ - 'A simple C++ Entity Component System'
 *  
 */

import java.io.PrintStream;
import java.lang.reflect.Type;
import java.util.BitSet;
import java.util.NoSuchElementException;

/**
 * The ECS management engine. Supplemented by the EntityManager, ComponentManager, and System Manager
 * These four classes provide the actual functionality of the ecs
 * <p>
 * See https://git.software.kauripeak.co.nz/BrychanD/JavaECS
 * for documentation and more information.
 */
public class Engine {
    // All internal functions write error messages to errorStream; which defaults to System.err
    // Can be set to a different PrintStream, to allow errors to be logged elsewhere
    protected static PrintStream errorStream = System.err;
    // References to the manager classes
    protected EntityManager entityManager;
    protected ComponentManager componentManager;
    protected SystemManager systemManager;

    /***************************************
     **        Engine Constructors        **
     ***************************************/
    public Engine(){
        entityManager = new EntityManager();
        componentManager = new ComponentManager();
        systemManager = new SystemManager();
        // In a non-ECS type manner (instead OO-like), provide an additional method to
        // add components to an entity: entity.addComponent().
        Entity.engineRef = this;
    }

    /**
     * Initialises the ECS with the specified value
     * @param maxEntities the maximum number of entities to allow
     */
    public Engine(int maxEntities){
        entityManager = new EntityManager(maxEntities);
        componentManager = new ComponentManager();
        systemManager = new SystemManager();
        Entity.engineRef = this;
    }

    /**
     * Attempts to resize the maximum number of entities
     * @param newSize the new maximum number of entities
     * @return true if the operation succeeded
     */
    public boolean resizeMaximum(int newSize){
        return entityManager.resize(newSize, systemManager, componentManager);
    }


    /***************************************
     **         Manage Entities           **
     ***************************************/

    /**
     * Create a new Entity (dequeues the first element from unusedEntities)
     * @return the next unused Entity, now set to used
     * @throws NoSuchElementException if there are no more entities available
     */
    public Entity createEntity() throws NoSuchElementException{
        try{
            Entity newEntity = entityManager.addEntity();
            return newEntity;
        }
        catch (NoSuchElementException e){
            // Catch the exception to log the error message
            writeErr("No more available entities");
            // Throw again to pass up to the original calling function
            throw e;
        }
    }

    /**
     * Signals each manager to remove the specified entity
     * @param entity the entity to destroy
     */
    public void destroyEntity(Entity entity){
        // Pass the registration list to the componentManager first, to avoid having to iterate through each component type
        componentManager.entityDestroyed(entity, entityManager.getRegistrations(entity));
        entityManager.removeEntity(entity);
        systemManager.entityDestroyed(entity);
    }

    /**
     * Gets the set maximum number of entities
     * @return an integer of the maximum number of entities
     */
    public Integer getMaxEntities(){
        return entityManager.getMaxSize();
    }

    /**
     * Gets the current number of entities
     * @return the number of currently active entities
     */
    public int getNumEntities(){
        return entityManager.getNumEntities();
    }

    /**
     * Checks if the entity is subscribed to the provided type
     * @param entity the entity to search
     * @param componentType the class type of the component to search
     * @return true if the entity is subscribed
     */
    public boolean entityHasComponent(Entity entity, Type componentType){
        return entityManager.getRegistrations(entity).get(componentManager.getComponentIndex(componentType));
    }

    /***************************************
     **         Manage Components         **
     ***************************************/

    /**
     * Registers the specified name in the component manager
     * @param type the name to register. Should be the component class name or a suitable name for primitive types
     * @return true if the component type was registered successfully
     */
    public boolean registerComponent(Type type){
        return componentManager.registerComponent(type);
    }

    /**
     * Adds an exisiting component to an exisiting entity
     * @param entity the entity to add the component to
     * @param componentType the class name of the component to add
     * @param component the actual component data
     * @return true if the compnent was added to the entity
     */
    public boolean addComponent(Entity entity, Type componentType, Object component){
        // Get the old registrations of the entity so that we can update only the relevant systems
        int componentIndex = componentManager.getComponentIndex(componentType);
        if (entityManager.registerComponent(componentIndex, entity))
        {
            systemManager.entityRegistrationsChanged(entity, entityManager.getRegistrations(entity));
            componentManager.addComponentToEntity(componentType, component, entity);
            return true;
        }
        else {
            writeErr("(" + componentType.getTypeName() + ")");
            return false;
        }
    }

    /**
     * Removes the component from the specified entity
     * @param entity the entity to remove the component from
     * @param componentType the class type of the component
     * @return true if the component was removed
     */
    public boolean removeComponent(Entity entity, Type componentType){
        if (entityManager.unregisterComponent(componentManager.getComponentIndex(componentType), entity))
        {
            componentManager.removeComponentFromEntity(componentType, entity);
            systemManager.entityRegistrationsChanged(entity, entityManager.getRegistrations(entity));
            return true;
        }
        else return false;
    }

    /**
     * Gets the actual data of the component associated to the entity.
     * Requires casting from Object to the known data type
     * @param entity the entity to retrieve the data for
     * @param componentType the class type of the component
     * @return the component data Object associated with the entity
     */
    public Object getComponentData(Entity entity, Type componentType){
        return componentManager.getComponent(componentType, entity);
    }

    /**
     * Gets the component index of the provided type
     * @param type the type to get the index of
     * @return the index of the component
     */
    public Integer getComponentIndex(Type type){
        return componentManager.getComponentIndex(type);
    }


    /****************************************
     **           Manage Systems           **
     ****************************************/
    
    /**
     * Registers the system to the SystemManager
     * @param systemType the type of the system
     * @param instance the instance of the system
     * @return true if successful
     */
    public boolean registerSystem(Type systemType, ECSSystem instance){
        return systemManager.registerSystem(systemType, instance);
    }
    
    /**
     * Sets the specified system's signature to the provided signature
     * @param system the class name of the system to set the signature of
     * @param signature the new signature data
     */
    public void setSystemSignature(Type system, BitSet signature){
        systemManager.setSystemRegistraions(system, signature);
    }

    
    /****************************************
     **    Helper Functions & Accessors    **
     ****************************************/

    /**
     * Writes an error message to the set PrintStream
     * <p>
     * <i>Default is System.err
     * @param message the message to write
     */
    static void writeErr(String message){
        errorStream.println(message);
    }

    /**
     * Writes an object to the set PrintStream
     * <p>
     * <i>Default System.err
     * @param obj the object to write
     */
    static void writeErr(Object obj){
        errorStream.println(obj);
    }

    /**
     * Sets the error PrintStream to the provided PrintStream
     * @param newErr the PrintStream to set as the write destination
     */
    public static void setErr(PrintStream newErr){
        errorStream = newErr;
    }

    /**
     * Gets the current error PrintStream
     * @return the current PrintStream that errors are written to
     */
    public static PrintStream getErr(){
        return errorStream;
    }

    /**
     * Gets a the current component manager
     * @return the current active component manager
     */
    public ComponentManager getComponentManager() {
        return componentManager;
    }

    /**
     * Gets a the current enitity manager
     * @return the current active enitity manager
     */
    public EntityManager getEntityManager() {
        return entityManager;
    }

    /**
     * Gets a the current systems manager
     * @return the current active systems manager
     */
    public SystemManager getSystemManager() {
        return systemManager;
    }
}