package nz.ac.massey.javaecs;

import java.lang.reflect.Type;

/**
 * Entity class. 
 * Whilst an entity is just an Integer, and using Integer values
 * would be more performant (no function call, primitive arrays and etc.), it is encapsulated to help
 * distinguish between entities and Integers. Also, in this form it may be extended to also include things
 * like names, grouping, specific ordering etc
 */
public class Entity {
    protected static Engine engineRef;

    private int value;

    public Entity(int value){
        this.value = value;
    }

    public int getValue() {
        return value;
    }

    /**
     * Returns the int value wrapped as an entity.
     * Used to provide a distinction between creating a new entity, and
     * using an int value we assume is a valid entity.
     * <p>
     * Functionally, this is no different to creating a new entity.
     * 
     * @param value the integer value to read as an Entity
     * @return an Entity object representing the ID value
     */
    public static Entity asEntity(int value){
        return new Entity(value);
    }

    @Override
    public boolean equals(Object obj) 
    {
        if (obj instanceof Entity){
            return (this.getValue() == ((Entity)obj).getValue());
        }
        else if (obj instanceof Integer){
            return (this.value == (Integer)obj);
        }
        else return false;
    }

    @Override
    public int hashCode(){
        return value; // each entity is an id; the value is implicitly unique, so use it as the hash code
    }

    /**
     * Adds the provided component to this entity.
     * <p>
     * <b>This function calls <i>Engine.addComponent(Entity, Type, Object)</i></b>, so
     * that should be used instead.
     * @deprecated This function is not ECS-like.
     * It is provided as an auxilliary method, that will be more
     * intuitive to those familiar with OO design
     * @param componentType the class type of the component to add
     * @param componentData the component data
     * @return true if successful
     */
    @Deprecated
    public boolean addComponent(Type componentType, Object componentData){
        return engineRef.addComponent(this, componentType, componentData);
    }

    /**
     * Removes the provided component from this entity.
     * <p>
     * <b>This function calls <i>Engine.removeComponent(Entity, Type)</i></b>, so
     * that should be used instead.
     * @deprecated This function is not ECS-like.
     * It is provided as an auxilliary method, that will be more
     * intuitive to those familiar with OO design
     * @param componentType the class type of the component to remove
     * @return true if successful
     */
    @Deprecated
    public boolean removeComponent(Type componentType){
        return engineRef.removeComponent(this, componentType);
    }

    /**
     * Gets the component data associated to this entity.
     * <p>
     * <b>This function calls <i>Engine.getComponentData()</i></b>, so
     * that should be used instead.
     * @deprecated This function is not ECS-like.
     * It is provided as an auxilliary method, that will be more
     * intuitive to those familiar with OO design
     * @param componentType the class type of the component to fetch data from
     * @return the data Object, which requires casting
     */
    @Deprecated
    public Object getComponent(Type componentType){
        return engineRef.getComponentData(this, componentType);
    }
}
